<?php

namespace FiloBlu\Flow\Model\Channel\In;

use Exception;
use FiloBlu\Flow\Helper\LoggerProvider;
use FiloBlu\Flow\Model\Channel\ConfigFactory;
use FiloBlu\Flow\Model\Imagesflow;
use FiloBlu\Flow\Model\ImagesflowFactory;
use Magento\Catalog\Model\Product;
use Magento\Catalog\Model\Product\Image\CacheFactory;
use Magento\Catalog\Model\ProductRepository;
use Magento\Eav\Model\ResourceModel\Entity\Attribute\Collection;
use Magento\Framework\App\Filesystem\DirectoryList;
use Magento\Framework\App\ObjectManager;
use Magento\Framework\App\ProductMetadataInterface;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\DataObject;
use Magento\Framework\DB\Adapter\AdapterInterface;
use Magento\Framework\Event\Manager;
use Magento\Framework\Exception\FileSystemException;
use Magento\Framework\Filesystem;
use Magento\Framework\Filesystem\Directory\WriteInterface;
use Magento\Framework\Model\ResourceModel\Iterator;
use Magento\Framework\Model\ResourceModel\IteratorFactory;
use Magento\Framework\ObjectManagerInterface;
use Magento\MediaStorage\Model\File\Uploader;
use Zend_Db;
use Zend_Db_Statement_Exception;

/**
 * Class Images
 * @package FiloBlu\Flow\Model\Channel\In
 */
class Images extends AbstractModel
{
    /**
     * @var string
     */
    const STATUS_RECEIVED = 'received';

    /**
     * @var string
     */
    const STATUS_PROCESSED = 'processed';

    /**
     * @var string
     */
    const STATUS_PARSED = 'parsed';

    /**
     * @var string
     */
    const STATUS_ERROR = 'error';

    /**
     * @var \FiloBlu\Flow\Helper\Images
     */
    protected $_imagesHelper;

    /**
     * @var ResourceConnection
     */
    protected $_resourceConnection;
    /**
     * @var AdapterInterface
     */
    protected $_connection;
    /**
     * @var ProductRepository
     */
    protected $_productRepository;
    /**
     * @var Filesystem
     */
    protected $_filesystem;
    /**
     * @var WriteInterface
     */
    protected $_mediaDirectory;
    /**
     * @var WriteInterface
     */
    protected $_varDirectory;
    /**
     * @var  CacheFactory
     */
    protected $_imageCacheFactory;
    /**
     * @var Product
     */
    protected $catalogModelProduct;
    /**
     * @var array
     */
    protected $_lookupAttributeSet = [];

    /**
     * @var
     */
    protected $channel_photo_config;
    /**
     * @var
     */
    protected $collection_of_images_to_parse;
    /**
     * @var
     */
    protected $array_of_data_for_csv;
    /**
     * @var string
     */
    protected $filename_separator = '_';
    /**
     * @var string
     */
    protected $image_alt_separator = ' ';
    /**
     * @var Manager
     */
    protected $eventManager;

    /**
     * @var int
     */
    protected $_errorsProcessingImages = 0;

    /**
     * @var int
     */
    protected $_imagesToProcess = 0;

    /**
     * @var string
     */
    protected $unique_image_code_pattern = 'unq';

    /**
     * @var string
     */
    protected $cache_folder = BP . '/pub/media/catalog/product/cache/';

    /**
     * Product entity link field
     *
     * @var string
     */
    protected $productEntityLinkField;

    /**
     * @var ImagesflowFactory
     */
    protected $_imageFlowFactory;

    /**
     * @var IteratorFactory
     */
    protected $iteratorFactory;

    /**
     * @var ProductMetadataInterface
     */
    protected $productMetadata;
    /**
     * @var string
     */
    protected $entityIdFieldName;

    /**
     *
     * TODO: check locks on import
     * TODO: vendor/magento/module-catalog-import-export/Model/Import/Product.php see how to use this class
     */
    /**
     * Images constructor.
     * @param LoggerProvider $loggerProvider
     * @param ConfigFactory $channelConfigFactory
     * @param \FiloBlu\Flow\Helper\Images $imagesHelper
     * @param Manager $eventManager
     * @param ResourceConnection $resourceConnection
     * @param ProductRepository $productRepository
     * @param Filesystem $filesystem
     * @param CacheFactory $imageCacheFactory
     * @param Product $catalogModelProduct
     * @param ImagesflowFactory $imageFlowFactory
     * @param IteratorFactory $iteratorFactory
     * @param ProductMetadataInterface $productMetadata
     * @param ObjectManagerInterface $objectManager
     * @throws FileSystemException
     */
    public function __construct(
        LoggerProvider $loggerProvider,
        ConfigFactory $channelConfigFactory,
        \FiloBlu\Flow\Helper\Images $imagesHelper,
        Manager $eventManager,
        ResourceConnection $resourceConnection,
        ProductRepository $productRepository,
        Filesystem $filesystem,
        CacheFactory $imageCacheFactory,
        Product $catalogModelProduct,
        ImagesflowFactory $imageFlowFactory,
        IteratorFactory $iteratorFactory,
        ProductMetadataInterface $productMetadata,
        ObjectManagerInterface $objectManager

    ) {
        parent::__construct($loggerProvider, $channelConfigFactory, $objectManager);

        $this->_imagesHelper = $imagesHelper;
        $this->eventManager = $eventManager;

        $this->_resourceConnection = $resourceConnection;

        $this->_productRepository = $productRepository;
        $this->_filesystem = $filesystem;
        $this->_mediaDirectory = $filesystem->getDirectoryWrite(DirectoryList::MEDIA);
        $this->_varDirectory = $filesystem->getDirectoryWrite(DirectoryList::VAR_DIR);
        $this->_imageCacheFactory = $imageCacheFactory;

        $this->catalogModelProduct = $catalogModelProduct;
        $this->_imageFlowFactory = $imageFlowFactory;
        $this->iteratorFactory = $iteratorFactory;
        $this->productMetadata = $productMetadata;
        $this->loadLookupTables();
    }

    /**
     * Load the lookups tables to speed up the whole process
     */
    protected function loadLookupTables()
    {
        $stmt = $this->getConnection()->query('SELECT a.attribute_set_id, b.* FROM eav_attribute_set AS a LEFT JOIN eav_entity_type AS b ON b.entity_type_id = a.entity_type_id WHERE attribute_set_id IN ( SELECT DISTINCT attribute_set_id FROM catalog_product_entity )');

        $results = $stmt->fetchAll(Zend_Db::FETCH_OBJ);

        foreach ($results as $item) {
            $item->attributesSet = $this->getAttributeSet($item->entity_type_id);
            $this->_lookupAttributeSet[$item->attribute_set_id] = $item;
        }
    }

    /**
     * @return AdapterInterface
     */
    protected function getConnection()
    {
        if ($this->_connection) {
            return $this->_connection;
        }

        return ($this->_connection = $this->_resourceConnection->getConnection());
    }

    /**
     * Retrieve the attribute set available from the entity type
     *
     * @param $entity_type_id
     * @return array
     * @throws Zend_Db_Statement_Exception
     */
    protected function getAttributeSet($entity_type_id)
    {
        $attributeSet = [];

        $stmt = $this->getConnection()->query(
            'SELECT * FROM eav_attribute WHERE entity_type_id = ?',
            [$entity_type_id]
        );

        $results = $stmt->fetchAll(Zend_Db::FETCH_OBJ);

        foreach ($results as $item) {
            $attributeSet[$item->attribute_code] = $item;
        }

        return $attributeSet;
    }

    /**
     * @return int
     */
    public function getProcessedImageErrorCount()
    {
        return $this->_errorsProcessingImages;
    }

    /**
     * @return int
     */
    public function getImageToProcessCount()
    {
        return $this->_imagesToProcess;
    }

    /**
     * @param $file
     * @return bool
     */
    public function insertData($file)
    {
        $meta_file_id = $file->getId();

        /** @var Imagesflow $model */
        $model = $this->_imageFlowFactory->create();

        // Loading images SKUs to process
        $collection = $model->getCollection();
        $collection->addFieldToSelect(['id', 'image_sku']);
        //$collection->addFieldToFilter('meta_processed', 0);
        $collection->addFieldToFilter('meta_file', $meta_file_id);
        $collection->addFieldToFilter('image_status', ['in' => [Imagesflow::STATUS_PARSED, Imagesflow::STATUS_ERROR]]);
        $collection->getSelect()->group('image_sku');
        $collection->load();

        $this->_imagesToProcess = $collection->count();

        /** @var Iterator $iterator */
        $iterator = $this->iteratorFactory->create();
        $iterator->walk($collection->getSelect(), [[$this, 'fastImageProcess']], ['meta_file_id' => $meta_file_id]);

        return !($this->_errorsProcessingImages > 0);
    }

    /**
     * Process image import directly query the db layer
     * @param $args
     * @return bool
     * @throws Zend_Db_Statement_Exception
     */
    public function fastImageProcess($args)
    {
        $entityIdFieldName = $this->getEntityField();

        /* Where Magento 2 looks for images */
        $catalogRootDirectory = $this->_mediaDirectory->getAbsolutePath();

        /* Retieve sku */
        $sku = $args['row']['image_sku'];
        $meta_file_id = $args['meta_file_id'];

        /* Find product by sku */

        $results = $this->getConnection()->query('SELECT * FROM catalog_product_entity WHERE sku = ?', [$sku]);

        $product = $results->fetch(Zend_Db::FETCH_OBJ);

        if ($product === null || $product === false) {
            $this->_logger->error("Unable to find product with SKU {$sku}");
            $this->_errorsProcessingImages++;
            return true;
        }

        /* TODO : use direct query instead. See https://blackfire.io/profiles/0b7ca6a2-ac40-41c1-b41d-1bfee4e7a8ac/graph*/

        /* Find incoming product images */

        $model = $this->_imageFlowFactory->create();

        $collection = $model->getCollection();
        // $collection->addFieldToFilter('meta_processed', 0);
        $collection->addFieldToFilter('meta_file', $meta_file_id);
        $collection->addFieldToFilter('image_status', ['in' => [Imagesflow::STATUS_PARSED, Imagesflow::STATUS_ERROR]]);
        $collection->addFieldToFilter('image_sku', $sku);
        //$collection->setOrder('image_position', 'ASC');
        $collection->load();

        $incomingImages = [];
        foreach ($collection as $imageFlow) {
            $incomingImages[Uploader::getCorrectFileName($imageFlow->getImageName())] = $imageFlow;
        }

        $id = (int)$product->attribute_set_id;
        $attributeSet = $this->_lookupAttributeSet[$id];

        $mediaGaleryAttribute = $attributeSet->attributesSet['media_gallery'];
        $mediaGaleryAttributeId = $mediaGaleryAttribute->attribute_id;

        /* Fix $catalogRootDirectory */

        $catalogRootDirectory .= str_replace('_', DIRECTORY_SEPARATOR, $attributeSet->entity_type_code);

        /* Find existing images in gallery */
        $existingImages = $this->getProductGalleryBySku($sku, $mediaGaleryAttributeId);

        foreach ($incomingImages as $name => $imageFlow) {
            try {
                $relativePath = Uploader::getDispretionPath($name) . DIRECTORY_SEPARATOR . $name;
                $to = $catalogRootDirectory . $relativePath;
                $from = $this->getImportImagePath($imageFlow);

                $folder = dirname($to) . DIRECTORY_SEPARATOR;

                if (false === is_dir($folder)) {
                    if (mkdir($folder, 0777, true) || is_dir($folder)) {
                        $this->_logger->error("Unable to create directory {$folder}");
                        $this->setProcecessedStatus($imageFlow, Imagesflow::STATUS_ERROR, "Unable to create directory {$folder}");
                        $this->_errorsProcessingImages++;
                        return true;
                    }
                }

                if (false === copy($from, $to)) {
                    $this->setProcecessedStatus($imageFlow, Imagesflow::STATUS_ERROR, "Unable to move {$from} -> {$to}");
                    $this->_errorsProcessingImages++;
                    continue;
                }

                /* If already esists image in gallery just copy/move the image file and update metas */
                if (array_key_exists($name, $existingImages)) {
                    $row = $existingImages[$name];
                    $this->updateGalleryValues($product->{$entityIdFieldName}, $row->value_id, $imageFlow);
                    $this->updateAttributes($product->{$entityIdFieldName}, $relativePath, $imageFlow, $attributeSet->attributesSet);
                } else {
                    $this->addImageToGallery($relativePath, $product->{$entityIdFieldName}, $mediaGaleryAttributeId, $imageFlow);
                    $this->addAttributes($product->{$entityIdFieldName}, $relativePath, $imageFlow, $attributeSet->attributesSet);
                }

                $this->setProcecessedStatus($imageFlow, Imagesflow::STATUS_PROCESSED);
            } catch (Exception $e) {
                $this->_logger->info("Importing image {$name} for product with SKU {$sku} failed. : " . $e->getMessage() . PHP_EOL . $e->getTraceAsString());
                $this->setProcecessedStatus($imageFlow, Imagesflow::STATUS_ERROR, $e->getMessage());
                $this->_errorsProcessingImages++;
            }
        }

        return true;
    }

    /**
     * @return string
     */
    protected function getEntityField()
    {
        if ($this->entityIdFieldName) {
            return $this->entityIdFieldName;
        }

        $edition = $this->productMetadata->getEdition();
        $this->entityIdFieldName = ($edition === 'Enterprise' || $edition === 'B2B') ? 'row_id' : 'entity_id';

        return $this->entityIdFieldName;
    }

    /**
     * Retrieve the whole gallery for the given product sku
     * @param $sku
     * @param $media_gallery_attribute_id
     *
     * @return array
     * @throws Zend_Db_Statement_Exception
     */
    protected function getProductGalleryBySku($sku, $media_gallery_attribute_id)
    {
        $gallery = [];
        $entityIdFieldName = $this->getEntityField();

        $results = $this->getConnection()->query(
            "SELECT CPEMGON.*, CPEMGVTE.{$entityIdFieldName} FROM catalog_product_entity AS CPE
             LEFT JOIN catalog_product_entity_media_gallery_value_to_entity AS CPEMGVTE ON CPE.{$entityIdFieldName} = CPEMGVTE.{$entityIdFieldName}
             LEFT JOIN catalog_product_entity_media_gallery AS CPEMGON ON CPEMGVTE.value_id = CPEMGON.value_id
             WHERE CPE.sku = ? AND CPEMGON.attribute_id = ?",
            [$sku, $media_gallery_attribute_id]
        );

        $images = $results->fetchAll(Zend_Db::FETCH_OBJ);

        foreach ($images as $image) {
            /* Remove the dipersion prefix */
            $gallery[basename($image->value)] = $image;
        }

        return $gallery;
    }

    /**
     * @param Imagesflow $imagesflow
     * @return string
     */
    protected function getImportImagePath(Imagesflow $imagesflow)
    {
        return $this->_varDirectory->getAbsolutePath() . DIRECTORY_SEPARATOR . $imagesflow->getImagePath() . DIRECTORY_SEPARATOR . $imagesflow->getImageName();
    }

    /**
     * @param Imagesflow $image
     * @param string $status
     * @param string $reason
     */
    protected function setProcecessedStatus(ImagesFlow $image, $status = Imagesflow::STATUS_PROCESSED, $reason = null)
    {
        $connection = $this->getConnection();
        $connection->update(
            $connection->getTableName('flow_imagesflow'),
            [
                'image_process_time' => date('Y-m-d H:i:s'),
                'meta_processed' => 1,
                'image_status' => $status,
                'log' => $reason
            ],
            $connection->quoteInto('id = ?', $image->getId())
        );
    }

    /**
     * @param $row_id
     * @param $value_id
     * @param Imagesflow $imageFlow
     */
    protected function updateGalleryValues($row_id, $value_id, Imagesflow $imageFlow)
    {
        $entityIdFieldName = $this->getEntityField();
        $connection = $this->getConnection();

        $valueTable = $connection->getTableName('catalog_product_entity_media_gallery_value');

        $where = $connection->quoteInto("value_id = ? AND {$entityIdFieldName} = ?", $value_id, $row_id);

        $connection->update(
            $valueTable,
            [
                'store_id' => 0,
                'label' => $imageFlow->getImageAlt(),
                'position' => $imageFlow->getImagePosition(),
                'disabled' => $imageFlow->getImageHide()
            ],
            $where
        );
    }

    /**
     * Update attributes. Allowed attributes to be set are defined in @param $row_id
     * @param $value
     * @param Imagesflow $imageFlow
     * @param array $attributesSet
     * @throws Zend_Db_Statement_Exception
     * @see \FiloBlu\Flow\Model\Channel\In\Images::$_imageAttributeCodes
     *
     */
    protected function updateAttributes($row_id, $value, Imagesflow $imageFlow, array $attributesSet)
    {
        $entityIdFieldName = $this->getEntityField();
        $connection = $this->getConnection();

        $assignableAttributes = explode('|', $imageFlow->getImageType());

        if (empty($assignableAttributes)) {
            return;
        }

        $attributesTable = $connection->getTableName('catalog_product_entity_varchar');

        $attributeIds = [];
        $toUpdate = [];
        $lookup = [];

        foreach ($assignableAttributes as $attributeCode) {
            if (!isset($attributesSet[$attributeCode])) {
                continue;
            }

            $id = $attributesSet[$attributeCode]->attribute_id;
            $attributeIds[] = (int)$id;
            $lookup[$id] = $attributeCode;
        }

        $stmt = $connection->query(
            $this->_connection
                ->select()
                ->from($attributesTable)
                ->where("{$entityIdFieldName} = ?", $row_id)
                ->where('store_id = ?', 0)
                ->where('attribute_id IN (?)', $attributeIds)
        );

        $existingAttributes = $stmt->fetchAll(Zend_Db::FETCH_OBJ);

        /* Find attributes to update and attributes to insert */

        foreach ($existingAttributes as $attribute) {
            $id = $attribute->attribute_id;
            $toUpdate[] = $lookup[$id];

            $where = $connection->quoteInto('value_id = ?', $attribute->value_id);

            $connection->update(
                $attributesTable,
                [
                    'store_id' => 0,
                    'attribute_id' => $id,
                    $entityIdFieldName => $row_id,
                    'value' => $value
                ],
                $where
            );
        }

        /* Add missing attributes */

        foreach (array_diff($assignableAttributes, $toUpdate) as $attributeCode) {
            if (!isset($attributesSet[$attributeCode])) {
                continue;
            }

            $attributeId = $attributesSet[$attributeCode]->attribute_id;

            $connection->insert($attributesTable, [
                'store_id' => 0,
                'attribute_id' => $attributeId,
                $entityIdFieldName => $row_id,
                'value' => $value
            ]);
        }
    }

    /**
     * @param $value
     * @param $row_id
     * @param $attribute_id
     * @param Imagesflow $imageFlow
     */
    protected function addImageToGallery($value, $row_id, $attribute_id, Imagesflow $imageFlow)
    {
        $entityIdFieldName = $this->getEntityField();
        $connection = $this->getConnection();

        $galleryTable = $connection->getTableName('catalog_product_entity_media_gallery');
        $relationTable = $connection->getTableName('catalog_product_entity_media_gallery_value_to_entity');
        $valueTable = $connection->getTableName('catalog_product_entity_media_gallery_value');

        /* Add new entry in the gallery */

        $connection->insert($galleryTable, [
            'value' => $value,
            'attribute_id' => $attribute_id,
            'media_type' => 'image',
            'disabled' => 0
        ]);

        /* Create relationship */
        $value_id = $connection->lastInsertId($galleryTable);

        $connection->insert($relationTable, [
            'value_id' => $value_id,
            $entityIdFieldName => $row_id
        ]);

        /* Add new value entry */

        $connection->insert($valueTable, [
            'value_id' => $value_id,
            $entityIdFieldName => $row_id,
            'store_id' => 0,
            'label' => $imageFlow->getImageAlt(),
            'position' => $imageFlow->getImagePosition(),
            'disabled' => $imageFlow->getImageHide()
        ]);
    }

    /**
     * Update attributes.
     *
     * @param $row_id
     * @param $value
     * @param Imagesflow $imageFlow
     * @param array $attributesSet
     */
    protected function addAttributes($row_id, $value, Imagesflow $imageFlow, array $attributesSet)
    {
        $connection = $this->getConnection();

        $entityIdFieldName = $this->getEntityField();

        $assignableAttributes = explode('|', $imageFlow->getImageType());

        if (empty($assignableAttributes)) {
            return;
        }

        $attributesTable = $connection->getTableName('catalog_product_entity_varchar');

        /* Lookup for attibute_code => attribute_id */

        foreach ($assignableAttributes as $attribute) {
            if (!isset($attributesSet[$attribute])) {
                continue;
            }

            $attributeId = $attributesSet[$attribute]->attribute_id;

            $connection->insertOnDuplicate($attributesTable, [
                'store_id' => 0,
                'attribute_id' => $attributeId,
                $entityIdFieldName => $row_id,
                'value' => $value
            ], ['value']);
        }
    }

    /**
     * @param $file
     * @return bool
     * @throws Exception
     */
    public function receiveFile($file)
    {
        if (!$this->_connector) {
            throw new Exception('no connector joined to channel');
        }

        $meta_file_id = $file->getId();

        $local_dir = $this->_connector->getLocalDir() . '/' . $meta_file_id;

        $this->_connector->setLocalDir($local_dir);

        $model = $this->_imageFlowFactory->create();

        $collection = $model->getCollection();
        $collection->addFieldToFilter('meta_processed', 0);
        $collection->addFieldToFilter('meta_file', $meta_file_id);
        $collection->addFieldToFilter('image_status', 'found');
        $collection->addFieldToFilter('image_queue_time', ['notnull' => true]);
        $collection->load();

        $result = true;

        foreach ($collection as $image) {
            try {

                // Is unique filename enabled?
                $unique_filename_pattern = null;
                if ($this->_imagesHelper->isUniqueUploadImagesActive()) {
                    $unique_filename_pattern = "_{$this->unique_image_code_pattern}" . $image->getId();
                }

                $received = $this->_connector->receiveResource($image->getData('image_name'), null, FTP_BINARY, true, $unique_filename_pattern);
                if ($received) {
                    $image->setData('image_receive_time', date('Y-m-d H:i:s'));
                    $image->setData('image_status', 'received');
                    if ($image->getData('image_name') !== $received) {
                        $log = $image->getData('log');
                        $log = trim("Image name fixed from '" . $image->getData('image_name') . "' to '{$received}'" . $log . "\n");
                        $image->setData('log', $log);
                        $image->setData('image_name', $received);
                    }
                    $image->save();
                } else {
                    $image->setData('image_status', 'found');
                    $log = $image->getData('log');
                    $log .= "Image not received\n";
                    $image->setData('log', $log);
                    $image->save();
                    $result = false;
                }
            } catch (Exception $e) {
                $log = $image->getData('log');
                $log .= $e->getMessage() . "\n";
                $image->setData('log', $log);
                $image->save();
                $result = false;
                //$this->_logger->log('FTP',$e->getMessage());
            }
        }

        //Can't close the connection or the next cycle will fail
        //$this->_connector->close();

        return $result;
    }

    /**
     * @param $file
     * @return bool
     * @throws Exception
     */
    public function parse($file)
    {
        if (!$this->_connector) {
            throw new Exception('no connector joined to channel');
        }

        $meta_file_id = $file->getId();

        //$local_dir = $this->_connector->getLocalDir().DIRECTORY_SEPARATOR.$meta_file_id;
        //$this->_connector->setLocalDir($local_dir);

        $channel_config = $this->_connector->getConfig();
        $local_dir = $channel_config['local_dir'] . DIRECTORY_SEPARATOR . $meta_file_id;

        //$this->directoryList->getPath('var');

        $csv_filename = $file->getName();
        $channel_name = $file->getChannel();
        $this->loadImageChannelConfig($channel_name);
        $this->filename_separator = $this->getImageSeparator();

        if (!$this->channel_photo_config) {
            throw new Exception('Wrong images configuration');
        }

        $this->loadImagesListToParse($meta_file_id);

        foreach ($this->collection_of_images_to_parse as $data) {
            $image_filename = $data->getImageName();
            //  $image_filename_no_ext = $this->removeExtensionFromFilename($image_filename);
            $loaded_data = $this->loadDataBasedOnFilename($image_filename, $channel_name);

            if ($loaded_data['sku'] && count($loaded_data) > 0 && $this->productExists($loaded_data['sku'])) {
                $data->setData('image_sku', $loaded_data['sku']);

                //Handling not numeric "L1", "L2", "L3", "L4" special positions (pattern inside filename)
                //Giving custom high order (order is mandatory, numeric and unique)

                switch ($loaded_data['position']) {
                    case 'l1':
                    case 'L1':
                        $loaded_data['position'] = 999;
                        break;
                    case 'l2':
                    case 'L2':
                        $loaded_data['position'] = 998;
                        break;
                    case 'l3':
                    case 'L3':
                        $loaded_data['position'] = 997;
                        break;
                    case 'l4':
                    case 'L4':
                        $loaded_data['position'] = 996;
                        break;
                    default:
                        #nothing to do
                }

                $data->setData('image_position', (int)$loaded_data['position']);

                if (isset($loaded_data['alt']) && $loaded_data['alt']) {
                    $data->setData('image_alt', $loaded_data['alt']);
                }

                $data->setData('image_type', implode('|', array_unique($loaded_data['image_type'])));
                $data->setData('image_hide', $loaded_data['image_hide']);
                $data->setData('image_path', $local_dir);
                $data->setData('image_status', 'parsed');
                $data->save();
            } else {
                $data->setData('image_status', 'error');
                $data->save();
                $log = $file->getData('log');
                $log .= "Can't parse image filename {$image_filename} or SKU does not exists, wrong format! {$csv_filename}\n";
                $file->setData('log', $log);
                $file->save();
                continue;
            }
        }

        // Handling replicate images for same simple products
        $duplicate_image_super_attribute = trim($this->_imagesHelper->isDuplicateImagesActive());

        if (!empty($duplicate_image_super_attribute)) {

            // Reload Images
            $model = $this->_imageFlowFactory->create();

            $collection = $model->getCollection();
            $collection->addFieldToFilter('meta_processed', 0);
            $collection->addFieldToFilter('meta_file', $meta_file_id);
            $collection->addFieldToFilter('image_status', 'parsed');
            $collection->load();

            // Rerunning the same collection
            foreach ($collection as $data) {
                if ($data->getData('image_status') === self::STATUS_PARSED) {
                    try {

                        /* TODO :  just a workaround */

                        $position = (int)$this->getFieldPosition($duplicate_image_super_attribute);
                        $image_filename_no_ext = $this->removeExtensionFromFilename($data->getImageName());
                        $filename_exploded = explode($this->filename_separator, $image_filename_no_ext);
                        $sku = $filename_exploded[$this->getFieldPosition('sku')];

                        //  $loaded_data = $this->loadDataBasedOnFilename($data->getImageName(), $channel_name);

                        /* END */

                        $sku_childs_array = $this->getSimpleFromConfigurableHavingAttribute($sku, $duplicate_image_super_attribute, $filename_exploded[$position]);

                        if ($sku_childs_array === false) {
                            continue;
                        }

                        if (isset($filename_exploded[$this->getFieldPosition('default')])) {
                            $data->setData('image_sku', $sku);
                            $data->setData('log', 'Image set as default');
                            $data->save();
                        }

                        // $sku_childs_array = $this->_imagesHelper->getAllChildsOfParentWithSameAttribute($data->getData('image_sku'), $duplicate_image_super_attribute);
                        // if (count($sku_childs_array) > 0) {
                        // For each children replicating the same image to be processed but on a different SKU

                        foreach ($sku_childs_array as $child_sku) {
                            $data->unsetData('id');
                            $data->setData('image_sku', $child_sku);
                            $data->setData('log', 'Image replicated');
                            $data->save();
                        }

                        // }
                    } catch (Exception $exception) {
                        $this->_logger->error('An error occurred while expanding: ' . $exception->getMessage());
                    }
                }
            }
        }

        return true;
    }

    /**
     * @param $channel_name
     */
    protected function loadImageChannelConfig($channel_name)
    {
        $this->channel_photo_config = null;
        if ($this->_parser->loadMap($channel_name)) {
            $config_mapping = $this->_parser->map->getFullMappedData();
            $this->channel_photo_config = $config_mapping;
        }
    }

    /**
     * @return string
     */
    protected function getImageSeparator()
    {
        $config = $this->channel_photo_config;

        $has_separator = (isset($config['options']['separator']) && trim($config['options']['separator']) !== '');
        return $has_separator ? $config['options']['separator'] : '_';
    }

    /**
     * @param $file_id
     */
    protected function loadImagesListToParse($file_id)
    {
        $model = $this->_imageFlowFactory->create();

        $collection = $model->getCollection();
        $collection->addFieldToFilter('meta_processed', 0);
        $collection->addFieldToFilter('meta_file', $file_id);
        $collection->addFieldToFilter('image_status', 'received');
        $collection->load();

        $this->collection_of_images_to_parse = $collection;
    }

    /**
     * @param $image_filename
     * @param null $channel_name
     * @return mixed
     * @throws Exception
     */
    protected function loadDataBasedOnFilename($image_filename, $channel_name = null)
    {
        $alt = [];
        $original_filename = $image_filename;

        // Remove extension
        $image_filename = $this->removeExtensionFromFilename($image_filename);

        $filename_exploded = explode($this->filename_separator, $image_filename);

        $sku_position = $this->getSkuFilenamePosition();

        $sku = $filename_exploded[$sku_position] ?? '_unknown_';

        $image_position_position = $this->GetImagePositionPosition();

        $position = $filename_exploded[$image_position_position] ?? 1;

        $image_alt_positions = $this->GetImageAltPositions();

        if (count($image_alt_positions) > 0) {
            foreach ($image_alt_positions as $pos) {
                // If ALT of image is equal to $this->unique_image_code_pattern then we have an error in the naming and
                // we need to skip that ALT value

                if (isset($filename_exploded[$pos]) && strpos($filename_exploded[$pos], $this->unique_image_code_pattern) !== 0) {
                    $alt[] = $filename_exploded[$pos];
                }
            }
            if (count($alt) > 0) {
                $alt = implode($this->image_alt_separator, $alt);
            }
        }

        if (empty($alt)) {
            $alt = '';
        }

        if (false !== ($defaultListing = $this->getFieldPosition('default', false)) && isset($filename_exploded[$defaultListing])) {
            $image_type = $this->getImageType($filename_exploded[$defaultListing]);
        } else {
            $image_type = $this->getImageType($position);
        }

        $image_type = array_merge($image_type, $this->getImageType($position));

        $image_hide = $this->getImageHide($position);

        $this->array_of_data_for_csv[$sku][$position]['filename'] = $original_filename;
        $this->array_of_data_for_csv[$sku][$position]['path'] = $this->_connector->getLocalDir();
        $this->array_of_data_for_csv[$sku][$position]['position'] = $position;
        if ($alt) {
            $this->array_of_data_for_csv[$sku][$position]['alt'] = $alt;
        }
        $this->array_of_data_for_csv[$sku][$position]['image_type'] = $image_type;
        $this->array_of_data_for_csv[$sku][$position]['image_hide'] = $image_hide;
        $this->array_of_data_for_csv[$sku][$position]['sku'] = $sku;

        // Start dispatch event to handle change of this data if needed
        $event_dispatch_data = new DataObject();
        $event_dispatch_data->setData('channel_name', $channel_name);
        $event_dispatch_data->setData('image_filename', $original_filename);
        $event_dispatch_data->setData('image_filename_noext', $image_filename);
        $event_dispatch_data->setData('filename_separator', $this->filename_separator);
        $event_dispatch_data->setData('channel_mapping', $this->channel_photo_config);
        $event_dispatch_data->setData('loaded_data', $this->array_of_data_for_csv[$sku][$position]);
        $this->eventManager->dispatch('flow_images_parse_load_data_based_on_filename_after', ['event_dispatch_data' => $event_dispatch_data]);
        // If the value is changed then it will be updated here
        $this->array_of_data_for_csv[$sku][$position] = $event_dispatch_data->getData('loaded_data');
        // End dispatch event, the data may have been changed here

        return $this->array_of_data_for_csv[$sku][$position];
    }

    /**
     * @param $string
     * @return bool|string
     */
    public function removeExtensionFromFilename($string)
    {
        $p = strrpos($string, '.');
        if ($p !== false) {
            $string = substr($string, 0, $p);
        }

        return $string;
    }

    /**
     * @return int|string
     */
    protected function getSkuFilenamePosition()
    {
        $config = $this->channel_photo_config;
        foreach ($config['fields'] as $key => $value) {
            if ($value['type'] === 'sku') {
                return $key;
            }
        }

        //Handle error missing SKU on config here, block all flow
    }

    /**
     * @return int|string
     */
    protected function GetImagePositionPosition()
    {
        $config = $this->channel_photo_config;
        foreach ($config['fields'] as $key => $value) {
            if ($value['type'] === 'position') {
                return $key;
            }
        }

        //Handle error missing SKU on config here, block all flow
    }

    /**
     * @return array
     */
    protected function GetImageAltPositions()
    {
        $config = $this->channel_photo_config;
        $positions = [];
        foreach ($config['fields'] as $key => $value) {
            if ($value['type'] === 'alt') {
                $positions[] = $key;
            }
        }

        return $positions;

        //Handle error missing SKU on config here, block all flow
    }

    /**
     * @param $fieldName
     * @param bool $throw | integer
     * @return bool|int|string
     * @throws Exception
     */
    protected function getFieldPosition($fieldName, $throw = true)
    {
        $config = $this->channel_photo_config;
        foreach ($config['fields'] as $key => $value) {
            if ($value['type'] === $fieldName) {
                return $key;
            }
        }

        if ($throw) {
            throw new Exception("Mapped field '{$fieldName}' does not exist in channel map");
        }

        return false;
    }

    /**
     * This will look into the json config of the image channels
     * searching which type of image is, based on the $position key inside the "types" node of the json
     *
     * @param $position
     *
     * @return array
     */
    protected function getImageType($position)
    {
        $position = trim($position);

        $config = $this->channel_photo_config;

        if (isset($config['types'][$position]['type'])) {
            return $config['types'][$position]['type'];
        }

        // Check if position is a number (can also be a string)
        // If is a number handle it as a number
        // This will handle positions as "001" "01" "1" (usually human errors)
        if ((int)$position > 0 && isset($config['types'][(int)$position]['type'])) {
            return $config['types'][(int)$position]['type'];
        }

        // Check again if maybe we have uppercases (usually human errors)
        $position = strtolower($position);
        return $config['types'][$position]['type'] ?? [];
    }

    /**
     * @param $position
     * @return int
     */
    protected function getImageHide($position)
    {
        $config = $this->channel_photo_config;

        if (isset($config['types'][$position]['hide']) && $config['types'][$position]['hide'] == "true") {
            return 1;
        }

        return 0;
    }

    /*
     * $filename is column 'value' from 'catalog_product_entity_media_gallery'
     */

    /**
     * @param string $sku
     * @return boolean
     */
    public function productExists($sku)
    {
        if ($this->catalogModelProduct->getIdBySku($sku)) {
            return true;
        }
        return false;
    }

    /**
     * @param $sku
     * @param $attributeId
     * @param $attributeValue
     * @return array|bool
     * @throws Exception
     */
    protected function getSimpleFromConfigurableHavingAttribute($sku, $attributeId, $attributeValue)
    {
        $connection = $this->getConnection();
        $entityIdFieldName = $this->getEntityField();

        $catalog = $connection->getTableName('catalog_product_entity');
        $superAttributesTable = $connection->getTableName('catalog_product_super_attribute');
        $eavAttributeTable = $connection->getTableName('eav_attribute');
        $eavAttributeOptionTable = $connection->getTableName('eav_attribute_option');
        $eavAttributeOptionValueTable = $connection->getTableName('eav_attribute_option_value');

        $superLinkTable = $connection->getTableName('catalog_product_super_link');

        /* 1] Find the configurable row_id */

        $configurables = $connection->fetchAll("SELECT {$entityIdFieldName} FROM {$catalog} WHERE sku = ? AND type_id = ?", [$sku, 'configurable'], Zend_Db::FETCH_OBJ);

        if (empty($configurables)) {
            return false;
        }

        $output = [];

        foreach ($configurables as $configurable) {
            $configurableId = $configurable->{$entityIdFieldName};

            /* 2] List super attributes */

            $superAttributes = [];
            $simpleIds = [];

            $attributes = $connection->fetchAll(
                "SELECT EA.attribute_id,EA.backend_type,EA.attribute_code 
FROM {$superAttributesTable} AS CPSA 
LEFT JOIN {$eavAttributeTable} AS EA ON EA.attribute_id = CPSA.attribute_id
WHERE CPSA.product_id = ?",
                [$configurableId],
                Zend_Db::FETCH_OBJ
            );

            foreach ($attributes as $attribute) {
                $superAttributes[] = $attribute->attribute_id;
            }

            /* 3] Find childs */

            $simples = $connection->fetchAll("SELECT {$entityIdFieldName}  FROM {$superLinkTable} AS CPSL LEFT JOIN {$catalog} AS CPE ON CPSL.product_id = CPE.entity_id WHERE CPSL.parent_id = ?", [$configurableId], Zend_Db::FETCH_OBJ);

            foreach ($simples as $simple) {
                $simpleIds[] = $simple->{$entityIdFieldName};
            }

            /* 4] Find childs matching attribute value */

            $sqlAttributesIds = implode(',', $superAttributes);
            $sqlSimpleIds = implode(',', $simpleIds);

            /**
             *
             * value_id |attribute_id |store_id |row_id |value |sku            |
             * ---------|-------------|---------|-------|------|---------------|
             * 6516450  |93           |0        |41888  |233   |BSC6A85ECG80TU |
             *
             */
            $sql = "SELECT CPEAV.{$entityIdFieldName}, CPEAV.attribute_id, CPE.sku
FROM catalog_product_entity_int AS CPEAV
LEFT JOIN {$catalog} AS CPE ON CPEAV.{$entityIdFieldName} = CPE.{$entityIdFieldName}
WHERE CPEAV.attribute_id IN ({$sqlAttributesIds}) AND CPEAV.`value` IN 
(
    SELECT EAO.option_id FROM {$eavAttributeOptionTable} AS EAO
    LEFT JOIN {$eavAttributeOptionValueTable} AS EAOV ON EAOV.option_id = EAO.option_id
    WHERE EAO.attribute_id IN ({$sqlAttributesIds}) AND (EAOV.`value` = ?) 
) 
AND CPEAV.{$entityIdFieldName} IN ({$sqlSimpleIds})";

            $matchedSimpleProducts = $connection->fetchAll($sql, [$attributeValue], Zend_Db::FETCH_OBJ);

            /* 5] check that matched simple products are all with same same attribute */
            $currentAttributeId = null;

            foreach ($matchedSimpleProducts as $simpleProduct) {
                $output[] = $simpleProduct->sku;
                if ($currentAttributeId === null) {
                    $currentAttributeId = $simpleProduct->attribute_id;
                    continue;
                }

                if ($currentAttributeId !== $simpleProduct->attribute_id) {
                    throw new Exception("The configurable product with sku '{$sku}' has simple products ({$sqlSimpleIds}) with the same attribute value '{$attributeValue}' for attributes with id ({$sqlAttributesIds}). Unable to resolve the ambiguity.");
                }
            }
        }

        return $output;
    }

    /**
     * @param $path
     * @return bool
     */
    protected function removeCatalogProductImageByPath($path)
    {
        $file = BP . "/pub/media/catalog/product{$path}";

        $connection = $this->getConnection();

        $query = "DELETE FROM catalog_product_entity_media_gallery WHERE value = '{$path}'";
        $result = $connection->query($query);
        $add_where = '';

        if (!$result) {
            $this->_logger->info("INSERT IMAGES: NOT Removed from catalog_product_entity_media_gallery {$path}");
            return false;
        }

        $this->_logger->info("INSERT IMAGES: Removed from catalog_product_entity_media_gallery {$path}");

        $ProductMediaImageAttributeIds = $this->getProductMediaImageAttributeIds();
        if ($ProductMediaImageAttributeIds) {
            $add_where = " AND attribute_id IN ({$ProductMediaImageAttributeIds}) ";
        }

        // Removing also the "photo tags" (base, small, thumbnail) or they will be left there
        $query = "DELETE FROM catalog_product_entity_varchar WHERE value = '{$path}' {$add_where}";
        $result = $connection->query($query);

        if (!$result) {
            $this->_logger->info("INSERT IMAGES: NOT Removed from catalog_product_entity_varchar {$path}");
            return false;
        }

        $this->_logger->info("INSERT IMAGES: Removed TAG from catalog_product_entity_varchar {$path}");

        if (is_file($file)) {
            unlink($file);
            $this->_logger->info("INSERT IMAGES: Removed file from disk {$file}");
        } else {
            $this->_logger->info("INSERT IMAGES: Removing file from disk {$file} not found, skipping");
        }

        $this->removeImageCacheFromImageFileName($path);

        return true;
    }

    /**
     * @return string of ids comma separated
     */
    protected function getProductMediaImageAttributeIds()
    {
        /** @var  $coll Collection */
        $coll = ObjectManager::getInstance()->create(Collection::class);
        $coll->addFieldToFilter('frontend_input', 'media_image');
        $attrAll = $coll->load()->getItems();
        $result = [];
        foreach ($attrAll as $attr) {
            $result[] = $attr->getAttributeId();
        }

        $return = false;
        if (count($result)) {
            $return = implode(',', $result);
        }

        return $return;
    }

    /**
     * @param null $filename
     * @return bool
     */
    protected function removeImageCacheFromImageFileName($filename = null)
    {
        if (!$filename || !$this->_imagesHelper->isClearImageCacheActive()) {
            return true;
        }

        // Array of all cache main folders
        try {
            $main_cache_image_paths = $this->getAllCacheImageMainPaths();
        } catch (Exception $e) {
            $this->_logger->info('ERROR IMAGES: removeImageCacheFromImageFileName ' . $e->getMessage());
            return false;
        }

        if (!isset($main_cache_image_paths) || !$main_cache_image_paths || count($main_cache_image_paths) === 0) {
            return false;
        }

        foreach ($main_cache_image_paths as $path) {

            // Safe check
            $path = rtrim($path, DIRECTORY_SEPARATOR);

            $full_path = $this->cache_folder . $path . $filename;

            if (file_exists($full_path)) {
                // Try to delete the cache
                unlink($full_path);
                $this->_logger->info("INSERT IMAGES: Removed cache image {$full_path}");
            }
        }
    }

    /**
     * @return array
     */
    public function getAllCacheImageMainPaths()
    {
        // Give me an array of all the folders inside the cache folder
        // Going deep 3
        $image_cache_paths_array = $this->_imagesHelper->foldersMap($this->cache_folder, 3);

        // Transform the array paths into string path checking also the md5 of the last folder
        $array_of_builded_paths = [];
        $this->_imagesHelper->collapse('', $image_cache_paths_array, $array_of_builded_paths);

        return $array_of_builded_paths;
    }
}
